// Geometric Tools, LLC
// Copyright (c) 1998-2010
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt
// http://www.geometrictools.com/License/Boost/LICENSE_1_0.txt
//
// File Version: 5.0.1 (2010/10/01)

#include "Wm5MathematicsPCH.h"
#include "Wm5DistPoint2Ellipse2.h"

namespace Wm5
{
//----------------------------------------------------------------------------
template <typename Real>
DistPoint2Ellipse2<Real>::DistPoint2Ellipse2 (const Vector2<Real>& point,
    const Ellipse2<Real>& ellipse)
    :
    mPoint(&point),
    mEllipse(&ellipse)
{
}
//----------------------------------------------------------------------------
template <typename Real>
const Vector2<Real>& DistPoint2Ellipse2<Real>::GetPoint () const
{
    return *mPoint;
}
//----------------------------------------------------------------------------
template <typename Real>
const Ellipse2<Real>& DistPoint2Ellipse2<Real>::GetEllipse () const
{
    return *mEllipse;
}
//----------------------------------------------------------------------------
template <typename Real>
Real DistPoint2Ellipse2<Real>::Get ()
{
    return Math<Real>::Sqrt(GetSquared());
}
//----------------------------------------------------------------------------
template <typename Real>
Real DistPoint2Ellipse2<Real>::GetSquared ()
{
    // Compute coordinates of point in ellipse coordinate system.
    Vector2<Real> diff = *mPoint - mEllipse->Center;
    Vector2<Real> epoint(diff.Dot(mEllipse->Axis[0]),
        diff.Dot(mEllipse->Axis[1]));

    const Real* extent = mEllipse->Extent;
    Real a2 = extent[0]*extent[0];
    Real b2 = extent[1]*extent[1];
    Real u2 = epoint.X()*epoint.X();
    Real v2 = epoint.Y()*epoint.Y();
    Real a2u2 = a2*u2, b2v2 = b2*v2;
    Real dx, dy, xDivA, yDivB, sqrDistance;
    Vector2<Real> closest;

    // Handle points near the coordinate axes.
    if (Math<Real>::FAbs(epoint.X()) <= Math<Real>::ZERO_TOLERANCE)
    {
        if (extent[0] >= extent[1] 
        ||  Math<Real>::FAbs(epoint.Y()) >= extent[1] - a2/extent[1])
        {
            closest.X() = (Real)0;
            if (epoint.Y() >= (Real)0)
            {
                closest.Y() = extent[1];
            }
            else
            {
                closest.Y() = -extent[1];
            }

            dy = closest.Y() - epoint.Y();
            sqrDistance = dy*dy;

            mClosestPoint0 = *mPoint;
            mClosestPoint1 = mEllipse->Center +
                closest.X()*mEllipse->Axis[0] +
                closest.Y()*mEllipse->Axis[1];
        }
        else
        {
            closest.Y() = b2*epoint.Y()/(b2 - a2);
            dy = closest.Y() - epoint.Y();
            yDivB = closest.Y()/extent[1];
            closest.X() = extent[0]*Math<Real>::Sqrt(
                Math<Real>::FAbs((Real)1 - yDivB*yDivB));

            sqrDistance = closest.X()*closest.X() + dy*dy;

            mClosestPoint0 = *mPoint;
            mClosestPoint1 = mEllipse->Center +
                closest.X()*mEllipse->Axis[0] +
                closest.Y()*mEllipse->Axis[1];
        }
        return sqrDistance;
    }

    if (Math<Real>::FAbs(epoint.Y()) <= Math<Real>::ZERO_TOLERANCE)
    {
        if (extent[1] >= extent[0]
        ||  Math<Real>::FAbs(epoint.X()) >= extent[0] - b2/extent[0])
        {
            closest.Y() = (Real)0;
            if (epoint.X() >= (Real)0)
            {
                closest.X() = extent[0];
            }
            else
            {
                closest.X() = -extent[0];
            }

            dx = closest.X() - epoint.X();
            sqrDistance = dx*dx;

            mClosestPoint0 = *mPoint;
            mClosestPoint1 = mEllipse->Center +
                mEllipse->Axis[0]*closest.X() +
                mEllipse->Axis[1]*closest.Y();
        }
        else
        {
            closest.X() = a2*epoint.X()/(a2 - b2);
            dx = closest.X() - epoint.X();
            xDivA = closest.X()/extent[0];
            closest.Y() = extent[1]*Math<Real>::Sqrt(
                Math<Real>::FAbs((Real)1 - xDivA*xDivA));

            sqrDistance = closest.Y()*closest.Y() + dx*dx;

            mClosestPoint0 = *mPoint;
            mClosestPoint1 = mEllipse->Center +
                closest.X()*mEllipse->Axis[0] +
                closest.Y()*mEllipse->Axis[1];
        }
        return sqrDistance;
    }

    // Initial guess for Newton's method.
    Real uRatio = epoint.X()/extent[0];
    Real vRatio = epoint.Y()/extent[1];
    Real t;
    if (uRatio*uRatio + vRatio*vRatio < (Real)1)
    {
        t = (Real)0;
    }
    else
    {
        Real maxExtent = extent[0];
        if (extent[1] > maxExtent)
        {
            maxExtent = extent[1];
        }

        t = maxExtent*epoint.Length();
    }

    // Apply Newton's method.
    const int maxIteration = 64;
    Real p = (Real)1;
    Real q = (Real)1;
    for (int i = 0; i < maxIteration; ++i)
    {
        p = t + a2;
        q = t + b2;
        Real p2 = p*p;
        Real q2 = q*q;
        Real r = p2*q2 - a2u2*q2 - b2v2*p2;
        if (Math<Real>::FAbs(r) < Math<Real>::ZERO_TOLERANCE)
        {
            break;
        }

        Real dr = ((Real)2)*(p*q*(p + q) - a2u2*q - b2v2*p);
        t -= r/dr;
    }

    closest.X() = a2*epoint.X()/p;
    closest.Y() = b2*epoint.Y()/q;
    diff = closest - epoint;
    sqrDistance = diff.SquaredLength();

    mClosestPoint0 = *mPoint;
    mClosestPoint1 = mEllipse->Center +
        closest.X()*mEllipse->Axis[0] +
        closest.Y()*mEllipse->Axis[1];

    return sqrDistance;
}
//----------------------------------------------------------------------------
template <typename Real>
Real DistPoint2Ellipse2<Real>::Get (Real t,
    const Vector2<Real>& velocity0, const Vector2<Real>& velocity1)
{
    Vector2<Real> movedPoint = *mPoint + t*velocity0;
    Vector2<Real> movedCenter = mEllipse->Center + t*velocity1;
    Ellipse2<Real> movedEllipse(movedCenter, mEllipse->Axis,
        mEllipse->Extent);
    return DistPoint2Ellipse2<Real>(movedPoint, movedEllipse).Get();
}
//----------------------------------------------------------------------------
template <typename Real>
Real DistPoint2Ellipse2<Real>::GetSquared (Real t,
    const Vector2<Real>& velocity0, const Vector2<Real>& velocity1)
{
    Vector2<Real> movedPoint = *mPoint + t*velocity0;
    Vector2<Real> movedCenter = mEllipse->Center + t*velocity1;
    Ellipse2<Real> movedEllipse(movedCenter, mEllipse->Axis,
        mEllipse->Extent);
    return DistPoint2Ellipse2<Real>(movedPoint, movedEllipse).GetSquared();
}
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// Explicit instantiation.
//----------------------------------------------------------------------------
template WM5_MATHEMATICS_ITEM
class DistPoint2Ellipse2<float>;

template WM5_MATHEMATICS_ITEM
class DistPoint2Ellipse2<double>;
//----------------------------------------------------------------------------
}
